#!/usr/bin/python
#
# Copyright Google 2007-2008, all rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import StringIO
import gdata
import gdata.service
import gdata.spreadsheet
import gdata.spreadsheet.service
import gdata.docs
import gdata.docs.service


"""Make the Google Documents API feel more like using a database.

This module contains a client and other classes which make working with the 
Google Documents List Data API and the Google Spreadsheets Data API look a
bit more like working with a heirarchical database. Using the DatabaseClient,
you can create or find spreadsheets and use them like a database, with 
worksheets representing tables and rows representing records.

Example Usage:
# Create a new database, a new table, and add records.
client = gdata.spreadsheet.text_db.DatabaseClient(username='jo@example.com', 
    password='12345')
database = client.CreateDatabase('My Text Database')
table = database.CreateTable('addresses', ['name','email',
    'phonenumber', 'mailingaddress'])
record = table.AddRecord({'name':'Bob', 'email':'bob@example.com', 
    'phonenumber':'555-555-1234', 'mailingaddress':'900 Imaginary St.'})

# Edit a record
record.content['email'] = 'bob2@example.com'
record.Push()

# Delete a table
table.Delete

Warnings: 
Care should be exercised when using this module on spreadsheets
which contain formulas. This module treats all rows as containing text and
updating a row will overwrite any formula with the output of the formula. 
The intended use case is to allow easy storage of text data in a spreadsheet.

  Error: Domain specific extension of Exception.
  BadCredentials: Error raised is username or password was incorrect.
  CaptchaRequired: Raised if a login attempt failed and a CAPTCHA challenge 
      was issued.
  DatabaseClient: Communicates with Google Docs APIs servers.
  Database: Represents a spreadsheet and interacts with tables.
  Table: Represents a worksheet and interacts with records.
  RecordResultSet: A list of records in a table.
  Record: Represents a row in a worksheet allows manipulation of text data.
"""


__author__ = 'api.jscudder (Jeffrey Scudder)'


class Error(Exception):
  pass


class BadCredentials(Error):
  pass


class CaptchaRequired(Error):
  pass


class DatabaseClient(object):
  """Allows creation and finding of Google Spreadsheets databases.

  The DatabaseClient simplifies the process of creating and finding Google 
  Spreadsheets and will talk to both the Google Spreadsheets API and the 
  Google Documents List API. 
  """

  def __init__(self, username=None, password=None):
    """Constructor for a Database Client. 
  
    If the username and password are present, the constructor  will contact
    the Google servers to authenticate.

    Args:
      username: str (optional) Example: jo@example.com
      password: str (optional)
    """
    self.__docs_client = gdata.docs.service.DocsService()
    self.__spreadsheets_client = (
        gdata.spreadsheet.service.SpreadsheetsService())
    self.SetCredentials(username, password)

  def SetCredentials(self, username, password):
    """Attempts to log in to Google APIs using the provided credentials.

    If the username or password are None, the client will not request auth 
    tokens.

    Args:
      username: str (optional) Example: jo@example.com
      password: str (optional)
    """
    self.__docs_client.email = username
    self.__docs_client.password = password
    self.__spreadsheets_client.email = username
    self.__spreadsheets_client.password = password
    if username and password:
      try:
        self.__docs_client.ProgrammaticLogin()
        self.__spreadsheets_client.ProgrammaticLogin()
      except gdata.service.CaptchaRequired:
        raise CaptchaRequired('Please visit https://www.google.com/accounts/'
                            'DisplayUnlockCaptcha to unlock your account.')
      except gdata.service.BadAuthentication:
        raise BadCredentials('Username or password incorrect.')
    
  def CreateDatabase(self, name):
    """Creates a new Google Spreadsheet with the desired name.

    Args:
      name: str The title for the spreadsheet.

    Returns:
      A Database instance representing the new spreadsheet. 
    """
    # Create a Google Spreadsheet to form the foundation of this database.
    # Spreadsheet is created by uploading a file to the Google Documents
    # List API.
    virtual_csv_file = StringIO.StringIO(',,,')
    virtual_media_source = gdata.MediaSource(file_handle=virtual_csv_file, content_type='text/csv', content_length=3)
    db_entry = self.__docs_client.UploadSpreadsheet(virtual_media_source, name)
    return Database(spreadsheet_entry=db_entry, database_client=self)

  def GetDatabases(self, spreadsheet_key=None, name=None):
    """Finds spreadsheets which have the unique key or title.

    If querying on the spreadsheet_key there will be at most one result, but
    searching by name could yield multiple results.

    Args:
      spreadsheet_key: str The unique key for the spreadsheet, this 
          usually in the the form 'pk23...We' or 'o23...423.12,,,3'.
      name: str The title of the spreadsheets.

    Returns:
      A list of Database objects representing the desired spreadsheets.
    """
    if spreadsheet_key:
      db_entry = self.__docs_client.GetDocumentListEntry(
          r'/feeds/documents/private/full/spreadsheet%3A' + spreadsheet_key)
      return [Database(spreadsheet_entry=db_entry, database_client=self)]
    else:
      title_query = gdata.docs.service.DocumentQuery()
      title_query['title'] = name
      db_feed = self.__docs_client.QueryDocumentListFeed(title_query.ToUri())
      matching_databases = []
      for entry in db_feed.entry:
        matching_databases.append(Database(spreadsheet_entry=entry, 
                                           database_client=self))
      return matching_databases
    
  def _GetDocsClient(self):
    return self.__docs_client

  def _GetSpreadsheetsClient(self):
    return self.__spreadsheets_client


class Database(object):
  """Provides interface to find and create tables.

  The database represents a Google Spreadsheet.
  """

  def __init__(self, spreadsheet_entry=None, database_client=None):
    """Constructor for a database object.

    Args:
      spreadsheet_entry: gdata.docs.DocumentListEntry The 
          Atom entry which represents the Google Spreadsheet. The
          spreadsheet's key is extracted from the entry and stored as a 
          member.
      database_client: DatabaseClient A client which can talk to the
          Google Spreadsheets servers to perform operations on worksheets
          within this spreadsheet.
    """
    self.entry = spreadsheet_entry
    if self.entry:
      id_parts = spreadsheet_entry.id.text.split('/')
      self.spreadsheet_key = id_parts[-1].replace('spreadsheet%3A', '')
    self.client = database_client

  def CreateTable(self, name, fields=None):
    """Add a new worksheet to this spreadsheet and fill in column names.

    Args:
      name: str The title of the new worksheet.
      fields: list of strings The column names which are placed in the
          first row of this worksheet. These names are converted into XML
          tags by the server. To avoid changes during the translation
          process I recommend using all lowercase alphabetic names. For
          example ['somelongname', 'theothername']

    Returns:
      Table representing the newly created worksheet.
    """
    worksheet = self.client._GetSpreadsheetsClient().AddWorksheet(title=name,
        row_count=1, col_count=len(fields), key=self.spreadsheet_key)
    return Table(name=name, worksheet_entry=worksheet, 
        database_client=self.client, 
        spreadsheet_key=self.spreadsheet_key, fields=fields)

  def GetTables(self, worksheet_id=None, name=None):
    """Searches for a worksheet with the specified ID or name.

    The list of results should have one table at most, or no results
    if the id or name were not found.

    Args:
      worksheet_id: str The ID of the worksheet, example: 'od6'
      name: str The title of the worksheet.

    Returns:
      A list of length 0 or 1 containing the desired Table. A list is returned
      to make this method feel like GetDatabases and GetRecords.
    """
    if worksheet_id:
      worksheet_entry = self.client._GetSpreadsheetsClient().GetWorksheetsFeed(
          self.spreadsheet_key, wksht_id=worksheet_id)
      return [Table(name=worksheet_entry.title.text, 
          worksheet_entry=worksheet_entry, database_client=self.client, 
          spreadsheet_key=self.spreadsheet_key)]
    else:
      matching_tables = []
      query = None
      if name:
        query = gdata.spreadsheet.service.DocumentQuery()
        query.title = name
 
      worksheet_feed = self.client._GetSpreadsheetsClient().GetWorksheetsFeed(
          self.spreadsheet_key, query=query)
      for entry in worksheet_feed.entry:
        matching_tables.append(Table(name=entry.title.text, 
            worksheet_entry=entry, database_client=self.client, 
            spreadsheet_key=self.spreadsheet_key))
      return matching_tables

  def Delete(self):
    """Deletes the entire database spreadsheet from Google Spreadsheets."""
    entry = self.client._GetDocsClient().Get(
        r'http://docs.google.com/feeds/documents/private/full/spreadsheet%3A' +
        self.spreadsheet_key)
    self.client._GetDocsClient().Delete(entry.GetEditLink().href)


class Table(object):

  def __init__(self, name=None, worksheet_entry=None, database_client=None, 
      spreadsheet_key=None, fields=None):
    self.name = name
    self.entry = worksheet_entry
    id_parts = worksheet_entry.id.text.split('/')
    self.worksheet_id = id_parts[-1]
    self.spreadsheet_key = spreadsheet_key
    self.client = database_client
    self.fields = fields or []
    if fields:
      self.SetFields(fields)

  def LookupFields(self):
    """Queries to find the column names in the first row of the worksheet.
    
    Useful when you have retrieved the table from the server and you don't 
    know the column names.
    """
    if self.entry:
      first_row_contents = []
      query = gdata.spreadsheet.service.CellQuery()
      query.max_row = '1'
      query.min_row = '1'
      feed = self.client._GetSpreadsheetsClient().GetCellsFeed(
          self.spreadsheet_key, wksht_id=self.worksheet_id, query=query)
      for entry in feed.entry:
        first_row_contents.append(entry.content.text)
      # Get the next set of cells if needed.
      next_link = feed.GetNextLink()
      while next_link:
        feed = self.client._GetSpreadsheetsClient().Get(next_link.href, 
            converter=gdata.spreadsheet.SpreadsheetsCellsFeedFromString)
        for entry in feed.entry:
          first_row_contents.append(entry.content.text)
        next_link = feed.GetNextLink()
      # Convert the contents of the cells to valid headers.
      self.fields = ConvertStringsToColumnHeaders(first_row_contents)
    
  def SetFields(self, fields):
    """Changes the contents of the cells in the first row of this worksheet.

    Args:
      fields: list of strings The names in the list comprise the
          first row of the worksheet. These names are converted into XML
          tags by the server. To avoid changes during the translation
          process I recommend using all lowercase alphabetic names. For
          example ['somelongname', 'theothername']
    """
    # TODO: If the table already had fields, we might want to clear out the,
    # current column headers.
    self.fields = fields
    i = 0
    for column_name in fields:
      i = i + 1
      # TODO: speed this up by using a batch request to update cells.
      self.client._GetSpreadsheetsClient().UpdateCell(1, i, column_name, 
          self.spreadsheet_key, self.worksheet_id)

  def Delete(self):
    """Deletes this worksheet from the spreadsheet."""
    worksheet = self.client._GetSpreadsheetsClient().GetWorksheetsFeed(
        self.spreadsheet_key, wksht_id=self.worksheet_id)
    self.client._GetSpreadsheetsClient().DeleteWorksheet(
        worksheet_entry=worksheet)

  def AddRecord(self, data):
    """Adds a new row to this worksheet.

    Args:
      data: dict of strings Mapping of string values to column names. 

    Returns:
      Record which represents this row of the spreadsheet.
    """
    new_row = self.client._GetSpreadsheetsClient().InsertRow(data, 
        self.spreadsheet_key, wksht_id=self.worksheet_id)
    return Record(content=data, row_entry=new_row, 
        spreadsheet_key=self.spreadsheet_key, worksheet_id=self.worksheet_id,
        database_client=self.client)

  def GetRecord(self, row_id=None, row_number=None):
    """Gets a single record from the worksheet based on row ID or number.
    
    Args:
      row_id: The ID for the individual row.
      row_number: str or int The position of the desired row. Numbering 
          begins at 1, which refers to the second row in the worksheet since
          the first row is used for column names.

    Returns:
      Record for the desired row.
    """
    if row_id:
      row_entry = self.client._GetSpreadsheetsClient().GetListFeed(
          self.spreadsheet_key, wksht_id=self.worksheet_id, row_id=row_id)
      return Record(content=None, row_entry=row_entry, 
           spreadsheet_key=self.spreadsheet_key, 
           worksheet_id=self.worksheet_id, database_client=self.client)
    else:
      row_query = gdata.spreadsheet.service.ListQuery()
      row_query.start_index = str(row_number)
      row_query.max_results = '1'
      row_feed = self.client._GetSpreadsheetsClient().GetListFeed(
          self.spreadsheet_key, wksht_id=self.worksheet_id, query=row_query)
      if len(row_feed.entry) >= 1:
        return Record(content=None, row_entry=row_feed.entry[0],
            spreadsheet_key=self.spreadsheet_key,
            worksheet_id=self.worksheet_id, database_client=self.client)
      else:
        return None

  def GetRecords(self, start_row, end_row):
    """Gets all rows between the start and end row numbers inclusive.

    Args:
      start_row: str or int
      end_row: str or int

    Returns:
      RecordResultSet for the desired rows.
    """
    start_row = int(start_row)
    end_row = int(end_row)
    max_rows = end_row - start_row + 1
    row_query = gdata.spreadsheet.service.ListQuery()
    row_query.start_index = str(start_row)
    row_query.max_results = str(max_rows)
    rows_feed = self.client._GetSpreadsheetsClient().GetListFeed(
        self.spreadsheet_key, wksht_id=self.worksheet_id, query=row_query)
    return RecordResultSet(rows_feed, self.client, self.spreadsheet_key,
        self.worksheet_id)

  def FindRecords(self, query_string):
    """Performs a query against the worksheet to find rows which match.

    For details on query string syntax see the section on sq under
    http://code.google.com/apis/spreadsheets/reference.html#list_Parameters

    Args:
      query_string: str Examples: 'name == john' to find all rows with john
          in the name column, '(cost < 19.50 and name != toy) or cost > 500'

    Returns:
      RecordResultSet with the first group of matches.
    """
    row_query = gdata.spreadsheet.service.ListQuery()
    row_query.sq = query_string
    matching_feed = self.client._GetSpreadsheetsClient().GetListFeed(
        self.spreadsheet_key, wksht_id=self.worksheet_id, query=row_query)
    return RecordResultSet(matching_feed, self.client, 
        self.spreadsheet_key, self.worksheet_id)


class RecordResultSet(list):
  """A collection of rows which allows fetching of the next set of results.

  The server may not send all rows in the requested range because there are
  too many. Using this result set you can access the first set of results
  as if it is a list, then get the next batch (if there are more results) by
  calling GetNext().
  """

  def __init__(self, feed, client, spreadsheet_key, worksheet_id):
    self.client = client
    self.spreadsheet_key = spreadsheet_key
    self.worksheet_id = worksheet_id
    self.feed = feed
    list(self)
    for entry in self.feed.entry:
      self.append(Record(content=None, row_entry=entry, 
          spreadsheet_key=spreadsheet_key, worksheet_id=worksheet_id,
          database_client=client))

  def GetNext(self):
    """Fetches the next batch of rows in the result set.

    Returns:
      A new RecordResultSet.
    """
    next_link = self.feed.GetNextLink()
    if next_link and next_link.href:
      new_feed = self.client._GetSpreadsheetsClient().Get(next_link.href, 
          converter=gdata.spreadsheet.SpreadsheetsListFeedFromString)
      return RecordResultSet(new_feed, self.client, self.spreadsheet_key,
          self.worksheet_id)


class Record(object):
  """Represents one row in a worksheet and provides a dictionary of values.

  Attributes:
    custom: dict Represents the contents of the row with cell values mapped
        to column headers.
  """

  def __init__(self, content=None, row_entry=None, spreadsheet_key=None, 
       worksheet_id=None, database_client=None):
    """Constructor for a record.
    
    Args:
      content: dict of strings Mapping of string values to column names.
      row_entry: gdata.spreadsheet.SpreadsheetsList The Atom entry 
          representing this row in the worksheet.
      spreadsheet_key: str The ID of the spreadsheet in which this row 
          belongs.
      worksheet_id: str The ID of the worksheet in which this row belongs.
      database_client: DatabaseClient The client which can be used to talk
          the Google Spreadsheets server to edit this row.
    """
    self.entry = row_entry
    self.spreadsheet_key = spreadsheet_key
    self.worksheet_id = worksheet_id
    if row_entry:
      self.row_id = row_entry.id.text.split('/')[-1]
    else:
      self.row_id = None
    self.client = database_client
    self.content = content or {}
    if not content:
      self.ExtractContentFromEntry(row_entry)

  def ExtractContentFromEntry(self, entry):
    """Populates the content and row_id based on content of the entry.

    This method is used in the Record's contructor.

    Args:
      entry: gdata.spreadsheet.SpreadsheetsList The Atom entry 
          representing this row in the worksheet.
    """
    self.content = {}
    if entry:
      self.row_id = entry.id.text.split('/')[-1]
      for label, custom in entry.custom.iteritems():
        self.content[label] = custom.text

  def Push(self):
    """Send the content of the record to spreadsheets to edit the row.

    All items in the content dictionary will be sent. Items which have been
    removed from the content may remain in the row. The content member
    of the record will not be modified so additional fields in the row
    might be absent from this local copy.
    """
    self.entry = self.client._GetSpreadsheetsClient().UpdateRow(self.entry, self.content)

  def Pull(self):
    """Query Google Spreadsheets to get the latest data from the server.

    Fetches the entry for this row and repopulates the content dictionary 
    with the data found in the row.
    """
    if self.row_id:
      self.entry = self.client._GetSpreadsheetsClient().GetListFeed(
          self.spreadsheet_key, wksht_id=self.worksheet_id, row_id=self.row_id)
    self.ExtractContentFromEntry(self.entry)

  def Delete(self):
    self.client._GetSpreadsheetsClient().DeleteRow(self.entry)


def ConvertStringsToColumnHeaders(proposed_headers):
  """Converts a list of strings to column names which spreadsheets accepts.

  When setting values in a record, the keys which represent column names must
  fit certain rules. They are all lower case, contain no spaces or special
  characters. If two columns have the same name after being sanitized, the 
  columns further to the right have _2, _3 _4, etc. appended to them.

  If there are column names which consist of all special characters, or if
  the column header is blank, an obfuscated value will be used for a column
  name. This method does not handle blank column names or column names with
  only special characters.
  """
  headers = []
  for input_string in proposed_headers:
    # TODO: probably a more efficient way to do this. Perhaps regex.
    sanitized = input_string.lower().replace('_', '').replace(
        ':', '').replace(' ', '')
    # When the same sanitized header appears multiple times in the first row
    # of a spreadsheet, _n is appended to the name to make it unique.
    header_count = headers.count(sanitized)
    if header_count > 0:
      headers.append('%s_%i' % (sanitized, header_count+1))
    else:
      headers.append(sanitized)
  return headers



#!/usr/bin/python
#
# Copyright (C) 2009 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains the data classes of the Dublin Core Metadata Initiative (DCMI) Extension"""


__author__ = 'j.s@google.com (Jeff Scudder)'


import atom.core


DC_TEMPLATE = '{http://purl.org/dc/terms/}%s'


class Creator(atom.core.XmlElement):
  """Entity primarily responsible for making the resource."""
  _qname = DC_TEMPLATE % 'creator'


class Date(atom.core.XmlElement):
  """Point or period of time associated with an event in the lifecycle of the resource."""
  _qname = DC_TEMPLATE % 'date'


class Description(atom.core.XmlElement):
  """Account of the resource."""
  _qname = DC_TEMPLATE % 'description'


class Format(atom.core.XmlElement):
  """File format, physical medium, or dimensions of the resource."""
  _qname = DC_TEMPLATE % 'format'


class Identifier(atom.core.XmlElement):
  """An unambiguous reference to the resource within a given context."""
  _qname = DC_TEMPLATE % 'identifier'


class Language(atom.core.XmlElement):
  """Language of the resource."""
  _qname = DC_TEMPLATE % 'language'


class Publisher(atom.core.XmlElement):
  """Entity responsible for making the resource available."""
  _qname = DC_TEMPLATE % 'publisher'


class Rights(atom.core.XmlElement):
  """Information about rights held in and over the resource."""
  _qname = DC_TEMPLATE % 'rights'


class Subject(atom.core.XmlElement):
  """Topic of the resource."""
  _qname = DC_TEMPLATE % 'subject'


class Title(atom.core.XmlElement):
  """Name given to the resource."""
  _qname = DC_TEMPLATE % 'title'





#!/usr/bin/python
#
# Copyright 2009 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains extensions to Atom objects used with Google Documents."""

__author__ = ('api.jfisher (Jeff Fisher), '
              'api.eric@google.com (Eric Bidelman)')

import atom
import gdata


DOCUMENTS_NAMESPACE = 'http://schemas.google.com/docs/2007'


class Scope(atom.AtomBase):
  """The DocList ACL scope element"""

  _tag = 'scope'
  _namespace = gdata.GACL_NAMESPACE
  _children = atom.AtomBase._children.copy()
  _attributes = atom.AtomBase._attributes.copy()
  _attributes['value'] = 'value'
  _attributes['type'] = 'type'

  def __init__(self, value=None, type=None, extension_elements=None,
               extension_attributes=None, text=None):
    self.value = value
    self.type = type
    self.text = text
    self.extension_elements = extension_elements or []
    self.extension_attributes = extension_attributes or {}


class Role(atom.AtomBase):
  """The DocList ACL role element"""

  _tag = 'role'
  _namespace = gdata.GACL_NAMESPACE
  _children = atom.AtomBase._children.copy()
  _attributes = atom.AtomBase._attributes.copy()
  _attributes['value'] = 'value'

  def __init__(self, value=None, extension_elements=None,
               extension_attributes=None, text=None):
    self.value = value
    self.text = text
    self.extension_elements = extension_elements or []
    self.extension_attributes = extension_attributes or {}


class FeedLink(atom.AtomBase):
  """The DocList gd:feedLink element"""

  _tag = 'feedLink'
  _namespace = gdata.GDATA_NAMESPACE
  _attributes = atom.AtomBase._attributes.copy()
  _attributes['rel'] = 'rel'
  _attributes['href'] = 'href'

  def __init__(self, href=None, rel=None, text=None, extension_elements=None,
               extension_attributes=None):
    self.href = href
    self.rel = rel
    atom.AtomBase.__init__(self, extension_elements=extension_elements,
                           extension_attributes=extension_attributes, text=text)


class ResourceId(atom.AtomBase):
  """The DocList gd:resourceId element"""

  _tag = 'resourceId'
  _namespace = gdata.GDATA_NAMESPACE
  _children = atom.AtomBase._children.copy()
  _attributes = atom.AtomBase._attributes.copy()
  _attributes['value'] = 'value'

  def __init__(self, value=None, extension_elements=None,
               extension_attributes=None, text=None):
    self.value = value
    self.text = text
    self.extension_elements = extension_elements or []
    self.extension_attributes = extension_attributes or {}


class LastModifiedBy(atom.Person):
  """The DocList gd:lastModifiedBy element"""

  _tag = 'lastModifiedBy'
  _namespace = gdata.GDATA_NAMESPACE


class LastViewed(atom.Person):
  """The DocList gd:lastViewed element"""

  _tag = 'lastViewed'
  _namespace = gdata.GDATA_NAMESPACE


class WritersCanInvite(atom.AtomBase):
  """The DocList docs:writersCanInvite element"""

  _tag = 'writersCanInvite'
  _namespace = DOCUMENTS_NAMESPACE
  _attributes = atom.AtomBase._attributes.copy()
  _attributes['value'] = 'value'


class DocumentListEntry(gdata.GDataEntry):
  """The Google Documents version of an Atom Entry"""

  _tag = gdata.GDataEntry._tag
  _namespace = atom.ATOM_NAMESPACE
  _children = gdata.GDataEntry._children.copy()
  _attributes = gdata.GDataEntry._attributes.copy()
  _children['{%s}feedLink' % gdata.GDATA_NAMESPACE] = ('feedLink', FeedLink)
  _children['{%s}resourceId' % gdata.GDATA_NAMESPACE] = ('resourceId',
                                                         ResourceId)
  _children['{%s}lastModifiedBy' % gdata.GDATA_NAMESPACE] = ('lastModifiedBy',
                                                             LastModifiedBy)
  _children['{%s}lastViewed' % gdata.GDATA_NAMESPACE] = ('lastViewed',
                                                         LastViewed)
  _children['{%s}writersCanInvite' % DOCUMENTS_NAMESPACE] = (
      'writersCanInvite', WritersCanInvite)

  def __init__(self, resourceId=None, feedLink=None, lastViewed=None,
               lastModifiedBy=None, writersCanInvite=None, author=None,
               category=None, content=None, atom_id=None, link=None,
               published=None, title=None, updated=None, text=None,
               extension_elements=None, extension_attributes=None):
    self.feedLink = feedLink
    self.lastViewed = lastViewed
    self.lastModifiedBy = lastModifiedBy
    self.resourceId = resourceId
    self.writersCanInvite = writersCanInvite
    gdata.GDataEntry.__init__(
        self, author=author, category=category, content=content,
        atom_id=atom_id, link=link, published=published, title=title,
        updated=updated, extension_elements=extension_elements,
        extension_attributes=extension_attributes, text=text)

  def GetAclLink(self):
    """Extracts the DocListEntry's <gd:feedLink>.

    Returns:
      A FeedLink object.
    """
    return self.feedLink

  def GetDocumentType(self):
    """Extracts the type of document from the DocListEntry.

    This method returns the type of document the DocListEntry
    represents.  Possible values are document, presentation,
    spreadsheet, folder, or pdf.

    Returns:
      A string representing the type of document.
    """
    if self.category:
      for category in self.category:
        if category.scheme == gdata.GDATA_NAMESPACE + '#kind':
          return category.label
    else:
      return None


def DocumentListEntryFromString(xml_string):
  """Converts an XML string into a DocumentListEntry object.

  Args:
    xml_string: string The XML describing a Document List feed entry.

  Returns:
    A DocumentListEntry object corresponding to the given XML.
  """
  return atom.CreateClassFromXMLString(DocumentListEntry, xml_string)


class DocumentListAclEntry(gdata.GDataEntry):
  """A DocList ACL Entry flavor of an Atom Entry"""

  _tag = gdata.GDataEntry._tag
  _namespace = gdata.GDataEntry._namespace
  _children = gdata.GDataEntry._children.copy()
  _attributes = gdata.GDataEntry._attributes.copy()
  _children['{%s}scope' % gdata.GACL_NAMESPACE] = ('scope', Scope)
  _children['{%s}role' % gdata.GACL_NAMESPACE] = ('role', Role)

  def __init__(self, category=None, atom_id=None, link=None,
               title=None, updated=None, scope=None, role=None,
               extension_elements=None, extension_attributes=None, text=None):
    gdata.GDataEntry.__init__(self, author=None, category=category,
                              content=None, atom_id=atom_id, link=link,
                              published=None, title=title,
                              updated=updated, text=None)
    self.scope = scope
    self.role = role


def DocumentListAclEntryFromString(xml_string):
  """Converts an XML string into a DocumentListAclEntry object.

  Args:
    xml_string: string The XML describing a Document List ACL feed entry.

  Returns:
    A DocumentListAclEntry object corresponding to the given XML.
  """
  return atom.CreateClassFromXMLString(DocumentListAclEntry, xml_string)


class DocumentListFeed(gdata.GDataFeed):
  """A feed containing a list of Google Documents Items"""

  _tag = gdata.GDataFeed._tag
  _namespace = atom.ATOM_NAMESPACE
  _children = gdata.GDataFeed._children.copy()
  _attributes = gdata.GDataFeed._attributes.copy()
  _children['{%s}entry' % atom.ATOM_NAMESPACE] = ('entry',
                                                  [DocumentListEntry])


def DocumentListFeedFromString(xml_string):
  """Converts an XML string into a DocumentListFeed object.

  Args:
    xml_string: string The XML describing a DocumentList feed.

  Returns:
    A DocumentListFeed object corresponding to the given XML.
  """
  return atom.CreateClassFromXMLString(DocumentListFeed, xml_string)


class DocumentListAclFeed(gdata.GDataFeed):
  """A DocList ACL feed flavor of a Atom feed"""

  _tag = gdata.GDataFeed._tag
  _namespace = atom.ATOM_NAMESPACE
  _children = gdata.GDataFeed._children.copy()
  _attributes = gdata.GDataFeed._attributes.copy()
  _children['{%s}entry' % atom.ATOM_NAMESPACE] = ('entry', 
                                                  [DocumentListAclEntry])


def DocumentListAclFeedFromString(xml_string):
  """Converts an XML string into a DocumentListAclFeed object.

  Args:
    xml_string: string The XML describing a DocumentList feed.

  Returns:
    A DocumentListFeed object corresponding to the given XML.
  """
  return atom.CreateClassFromXMLString(DocumentListAclFeed, xml_string)



#!/usr/bin/python
#
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DocsClient simplifies interactions with the Documents List API."""

__author__ = 'vicfryzel@google.com (Vic Fryzel)'

import copy
import mimetypes
import re
import urllib
import atom.data
import atom.http_core
import gdata.client
import gdata.docs.data
import gdata.gauth


# Feed URIs that are given by the API, but cannot be obtained without
# making a mostly unnecessary HTTP request.
RESOURCE_FEED_URI = '/feeds/default/private/full'
RESOURCE_SELF_LINK_TEMPLATE = RESOURCE_FEED_URI + '/%s'
RESOURCE_UPLOAD_URI = '/feeds/upload/create-session/default/private/full'
ARCHIVE_FEED_URI = '/feeds/default/private/archive'
METADATA_URI = '/feeds/metadata/default'
CHANGE_FEED_URI = '/feeds/default/private/changes'


class DocsClient(gdata.client.GDClient):
  """Client for all features of the Google Documents List API."""

  host = 'docs.google.com'
  api_version = '3.0'
  auth_service = 'writely'
  alt_auth_service = 'wise'
  alt_auth_token = None
  auth_scopes = gdata.gauth.AUTH_SCOPES['writely']
  ssl = True

  def request(self, method=None, uri=None, **kwargs):
    """Add support for imitating other users via 2-Legged OAuth.

    Args:
      uri: (optional) URI of the request in which to replace default with
          self.xoauth_requestor_id.
    Returns:
      Result of super(DocsClient, self).request().
    """
    if self.xoauth_requestor_id is not None and uri is not None:
      if isinstance(uri, (str, unicode)):
        uri = atom.http_core.Uri.parse_uri(uri)
      uri.path.replace('/default', '/%s' % self.xoauth_requestor_id)
    return super(DocsClient, self).request(method=method, uri=uri, **kwargs)

  Request = request

  def get_metadata(self, **kwargs):
    """Retrieves the metadata of a user account.

    Args:
      kwargs: Other parameters to pass to self.get_entry().

    Returns:
      gdata.docs.data.Metadata representing metadata of user's account.
    """
    return self.get_entry(
        METADATA_URI, desired_class=gdata.docs.data.Metadata, **kwargs)

  GetMetadata = get_metadata

  def get_changes(self, changestamp=None, max_results=None, show_root=None,
                  **kwargs):
    """Retrieves changes to a user's documents list.

    Args:
      changestamp: (optional) String changestamp value to query since.
          If provided, returned changes will have a changestamp larger than
          the given one.
      max_results: (optional) Number of results to fetch.  API will limit
          this number to 100 at most.
      show_root: (optional) True to include indications if a resource is in
          the root collection.
      kwargs: Other parameters to pass to self.get_feed().

    Returns:
      gdata.docs.data.ChangeFeed.
    """
    uri = atom.http_core.Uri.parse_uri(CHANGE_FEED_URI)

    if changestamp is not None:
      uri.query['start-index'] = changestamp
    if max_results is not None:
      uri.query['max-results'] = max_results
    if show_root is not None:
      uri.query['showroot'] = str(show_root).lower()

    return self.get_feed(
        uri, desired_class=gdata.docs.data.ChangeFeed, **kwargs)

  GetChanges = get_changes

  def get_resources(self, uri=None, limit=None, show_root=None, **kwargs):
    """Retrieves the resources in a user's docslist, or the given URI.

    Args:
      uri: (optional) URI to query for resources.  If None, then
          gdata.docs.client.DocsClient.RESOURCE_FEED_URI is used, which will
          query for all non-collections.
      limit: int (optional) A maximum cap for the number of results to
          return in the feed. By default, the API returns a maximum of 100
          per page. Thus, if you set limit=5000, you will get <= 5000
          documents (guarenteed no more than 5000), and will need to follow the
          feed's next links (feed.GetNextLink()) to the rest. See
          get_everything(). Similarly, if you set limit=50, only <= 50
          documents are returned. Note: if the max-results parameter is set in
          the uri parameter, it is chosen over a value set for limit.
      show_root: (optional) True to include indications if a resource is in
          the root collection.
      kwargs: Other parameters to pass to self.get_feed().

    Returns:
      gdata.docs.data.ResourceFeed feed.
    """
    if uri is None:
      uri = RESOURCE_FEED_URI

    if isinstance(uri, basestring):
      uri = atom.http_core.Uri.parse_uri(uri)

    # Add max-results param if it wasn't included in the uri.
    if limit is not None and not 'max-results' in uri.query:
      uri.query['max-results'] = limit
    if show_root is not None:
      uri.query['showroot'] = str(show_root).lower()


    return self.get_feed(uri, desired_class=gdata.docs.data.ResourceFeed,
                         **kwargs)

  GetResources = get_resources

  def get_all_resources(self, uri=None, show_root=None, **kwargs):
    """Retrieves all of a user's non-collections or everything at the given URI.

    Folders are not included in this by default.  Pass in a custom URI to
    include collections in your query.  The DocsQuery class is an easy way to
    generate such a URI.

    This method makes multiple HTTP requests (by following the feed's next
    links) in order to fetch the user's entire document list.

    Args:
      uri: (optional) URI to query the doclist feed with. If None, then use
          DocsClient.RESOURCE_FEED_URI, which will retrieve all
          non-collections.
      show_root: (optional) True to include indications if a resource is in
          the root collection.
      kwargs: Other parameters to pass to self.GetResources().

    Returns:
      List of gdata.docs.data.Resource objects representing the retrieved
      entries.
    """
    if uri is None:
      uri = RESOURCE_FEED_URI

    if isinstance(uri, basestring):
      uri = atom.http_core.Uri.parse_uri(uri)

    if show_root is not None:
      uri.query['showroot'] = str(show_root).lower()

    feed = self.GetResources(uri=uri, **kwargs)
    entries = feed.entry

    while feed.GetNextLink() is not None:
      feed = self.GetResources(feed.GetNextLink().href, **kwargs)
      entries.extend(feed.entry)

    return entries

  GetAllResources = get_all_resources

  def get_resource(self, entry, **kwargs):
    """Retrieves a resource again given its entry.

    Args:
      entry: gdata.docs.data.Resource to fetch and return.
      kwargs: Other args to pass to GetResourceBySelfLink().
    Returns:
      gdata.docs.data.Resource representing retrieved resource.
    """
    return self.GetResourceBySelfLink(entry.GetSelfLink().href, **kwargs)

  GetResource = get_resource

  def get_resource_by_id(self, resource_id, **kwargs):
    """Retrieves a resource again given its resource ID.

    Args:
      resource_id: Typed or untyped resource ID of a resource.
      kwargs: Other args to pass to GetResourceBySelfLink().
    Returns:
      gdata.docs.data.Resource representing retrieved resource.
    """
    return self.GetResourceBySelfLink(
        RESOURCE_SELF_LINK_TEMPLATE % resource_id, **kwargs)

  GetResourceById = get_resource_by_id

  def get_resource_by_self_link(self, uri, etag=None, show_root=None,
                                **kwargs):
    """Retrieves a particular resource by its self link.

    Args:
      uri: str URI at which to query for given resource.  This can be found
          using entry.GetSelfLink().
      etag: str (optional) The document/item's etag value to be used in a
          conditional GET. See http://code.google.com/apis/documents/docs/3.0/
          developers_guide_protocol.html#RetrievingCached.
      show_root: (optional) True to include indications if a resource is in
          the root collection.
      kwargs: Other parameters to pass to self.get_entry().

    Returns:
      gdata.docs.data.Resource representing the retrieved resource.
    """
    if isinstance(uri, basestring):
      uri = atom.http_core.Uri.parse_uri(uri)
    if show_root is not None:
      uri.query['showroot'] = str(show_root).lower()
    return self.get_entry(
        uri , etag=etag, desired_class=gdata.docs.data.Resource, **kwargs)

  GetResourceBySelfLink = get_resource_by_self_link

  def get_resource_acl(self, entry, **kwargs):
    """Retrieves the ACL sharing permissions for the given entry.

    Args:
      entry: gdata.docs.data.Resource for which to get ACL.
      kwargs: Other parameters to pass to self.get_feed().

    Returns:
      gdata.docs.data.AclFeed representing the resource's ACL.
    """
    self._check_entry_is_resource(entry)
    return self.get_feed(entry.GetAclFeedLink().href,
                         desired_class=gdata.docs.data.AclFeed, **kwargs)

  GetResourceAcl = get_resource_acl

  def create_resource(self, entry, media=None, collection=None,
                      create_uri=None, **kwargs):
    """Creates new entries in Google Docs, and uploads their contents.

    Args:
      entry: gdata.docs.data.Resource representing initial version
          of entry being created. If media is also provided, the entry will
          first be created with the given metadata and content.
      media: (optional) gdata.data.MediaSource containing the file to be
          uploaded.
      collection: (optional) gdata.docs.data.Resource representing a collection
          in which this new entry should be created. If provided along
          with create_uri, create_uri will win (e.g. entry will be created at
          create_uri, not necessarily in given collection).
      create_uri: (optional) String URI at which to create the given entry. If
          collection, media and create_uri are None, use
          gdata.docs.client.RESOURCE_FEED_URI.  If collection and create_uri are
          None, use gdata.docs.client.RESOURCE_UPLOAD_URI.  If collection and
          media are not None,
          collection.GetResumableCreateMediaLink() is used,
          with the collection's resource ID substituted in.
      kwargs: Other parameters to pass to self.post() and self.update().

    Returns:
      gdata.docs.data.Resource containing information about new entry.
    """
    if media is not None:
      if create_uri is None and collection is not None:
        create_uri = collection.GetResumableCreateMediaLink().href
      elif create_uri is None:
        create_uri = RESOURCE_UPLOAD_URI
      uploader = gdata.client.ResumableUploader(
          self, media.file_handle, media.content_type, media.content_length,
          desired_class=gdata.docs.data.Resource)
      return uploader.upload_file(create_uri, entry, **kwargs)
    else:
      if create_uri is None and collection is not None:
        create_uri = collection.content.src
      elif create_uri is None:
        create_uri = RESOURCE_FEED_URI
      return self.post(
          entry, create_uri, desired_class=gdata.docs.data.Resource, **kwargs)

  CreateResource = create_resource

  def update_resource(self, entry, media=None, update_metadata=True,
                      new_revision=False, **kwargs):
    """Updates an entry in Google Docs with new metadata and/or new data.

    Args:
      entry: Entry to update. Make any metadata changes to this entry.
      media: (optional) gdata.data.MediaSource object containing the file with
          which to replace the entry's data.
      update_metadata: (optional) True to update the metadata from the entry
          itself.  You might set this to False to only update an entry's
          file content, and not its metadata.
      new_revision: (optional) True to create a new revision with this update,
          False otherwise.
      kwargs: Other parameters to pass to self.post().

    Returns:
      gdata.docs.data.Resource representing the updated entry.
    """

    uri_params = {}
    if new_revision:
      uri_params['new-revision'] = 'true'

    if update_metadata and media is None:
      uri = atom.http_core.parse_uri(entry.GetEditLink().href)
      uri.query.update(uri_params)
      return super(DocsClient, self).update(entry, **kwargs)
    else:
      uploader = gdata.client.ResumableUploader(
          self, media.file_handle, media.content_type, media.content_length,
          desired_class=gdata.docs.data.Resource)
      return uploader.UpdateFile(entry_or_resumable_edit_link=entry,
                                 update_metadata=update_metadata,
                                 uri_params=uri_params, **kwargs)

  UpdateResource = update_resource

  def download_resource(self, entry, file_path, extra_params=None, **kwargs):
    """Downloads the contents of the given entry to disk.

    Note: to download a file in memory, use the DownloadResourceToMemory()
    method.

    Args:
      entry: gdata.docs.data.Resource whose contents to fetch.
      file_path: str Full path to which to save file.
      extra_params: dict (optional) A map of any further parameters to control
          how the document is downloaded/exported. For example, exporting a
          spreadsheet as a .csv: extra_params={'gid': 0, 'exportFormat': 'csv'}
      kwargs: Other parameters to pass to self._download_file().

    Raises:
      gdata.client.RequestError if the download URL is malformed or the server's
      response was not successful.
    """
    self._check_entry_is_not_collection(entry)
    self._check_entry_has_content(entry)
    uri = self._get_download_uri(entry.content.src, extra_params)
    self._download_file(uri, file_path, **kwargs)

  DownloadResource = download_resource

  def download_resource_to_memory(self, entry, extra_params=None, **kwargs):
    """Returns the contents of the given entry.

    Args:
      entry: gdata.docs.data.Resource whose contents to fetch.
      extra_params: dict (optional) A map of any further parameters to control
          how the document is downloaded/exported. For example, exporting a
          spreadsheet as a .csv: extra_params={'gid': 0, 'exportFormat': 'csv'}
      kwargs: Other parameters to pass to self._get_content().

    Returns:
      Content of given resource after being downloaded.

    Raises:
      gdata.client.RequestError if the download URL is malformed or the server's
      response was not successful.
    """
    self._check_entry_is_not_collection(entry)
    self._check_entry_has_content(entry)
    uri = self._get_download_uri(entry.content.src, extra_params)
    return self._get_content(uri, **kwargs)

  DownloadResourceToMemory = download_resource_to_memory

  def _get_download_uri(self, base_uri, extra_params=None):
    uri = base_uri.replace('&amp;', '&')
    if extra_params is not None:
      if 'exportFormat' in extra_params and '/Export?' not in uri:
        raise gdata.client.Error, ('This entry type cannot be exported '
                                   'as a different format.')

      if 'gid' in extra_params and uri.find('spreadsheets') == -1:
        raise gdata.client.Error, 'gid param is not valid for this resource type.'

      uri += '&' + urllib.urlencode(extra_params)
    return uri

  def _get_content(self, uri, extra_params=None, auth_token=None, **kwargs):
    """Fetches the given resource's content.

    This method is useful for downloading/exporting a file within enviornments
    like Google App Engine, where the user may not have the ability to write
    the file to a local disk.

    Be warned, this method will use as much memory as needed to store the
    fetched content.  This could cause issues in your environment or app. This
    is only different from Download() in that you will probably retain an
    open reference to the data returned from this method, where as the data
    from Download() will be immediately written to disk and the memory
    freed.  This client library currently doesn't support reading server
    responses into a buffer or yielding an open file pointer to responses.

    Args:
      entry: Resource to fetch.
      extra_params: dict (optional) A map of any further parameters to control
          how the document is downloaded/exported. For example, exporting a
          spreadsheet as a .csv: extra_params={'gid': 0, 'exportFormat': 'csv'}
      auth_token: (optional) gdata.gauth.ClientLoginToken, AuthSubToken, or
          OAuthToken which authorizes this client to edit the user's data.
      kwargs: Other parameters to pass to self.request().

    Returns:
      The binary file content.

    Raises:
      gdata.client.RequestError: on error response from server.
    """
    server_response = None
    token = auth_token
    if 'spreadsheets' in uri and token is None \
        and self.alt_auth_token is not None:
      token = self.alt_auth_token
    server_response = self.request(
        'GET', uri, auth_token=token, **kwargs)
    if server_response.status != 200:
      raise gdata.client.RequestError, {'status': server_response.status,
                                        'reason': server_response.reason,
                                        'body': server_response.read()}
    return server_response.read()

  def _download_file(self, uri, file_path, **kwargs):
    """Downloads a file to disk from the specified URI.

    Note: to download a file in memory, use the GetContent() method.

    Args:
      uri: str The full URL to download the file from.
      file_path: str The full path to save the file to.
      kwargs: Other parameters to pass to self.get_content().

    Raises:
      gdata.client.RequestError: on error response from server.
    """
    f = open(file_path, 'wb')
    try:
      f.write(self._get_content(uri, **kwargs))
    except gdata.client.RequestError, e:
      f.close()
      raise e
    f.flush()
    f.close()

  _DownloadFile = _download_file

  def copy_resource(self, entry, title, **kwargs):
    """Copies the given entry to a new entry with the given title.

    Note: Files do not support this feature.

    Args:
      entry: gdata.docs.data.Resource to copy.
      title: String title for the new entry.
      kwargs: Other parameters to pass to self.post().

    Returns:
      gdata.docs.data.Resource representing duplicated resource.
    """
    self._check_entry_is_resource(entry)
    new_entry = gdata.docs.data.Resource(
        title=atom.data.Title(text=title),
        id=atom.data.Id(text=entry.GetSelfLink().href))
    return self.post(new_entry, RESOURCE_FEED_URI, **kwargs)

  CopyResource = copy_resource

  def move_resource(self, entry, collection=None, keep_in_collections=False,
                    **kwargs):
    """Moves an item into a different collection (or out of all collections).

    Args:
      entry: gdata.docs.data.Resource to move.
      collection: gdata.docs.data.Resource (optional) An object representing
          the destination collection. If None, set keep_in_collections to
          False to remove the item from all collections.
      keep_in_collections: boolean (optional) If True, the given entry
          is not removed from any existing collections it is already in.
      kwargs: Other parameters to pass to self.post().

    Returns:
      gdata.docs.data.Resource of the moved entry.
    """
    self._check_entry_is_resource(entry)

    # Remove the item from any collections it is already in.
    if not keep_in_collections:
      for current_collection in entry.InCollections():
        uri = '%s/contents/%s' % (
            current_collection.href,
            urllib.quote(entry.resource_id.text))
        self.delete(uri, force=True)

    if collection is not None:
      self._check_entry_is_collection(collection)
      entry = self.post(entry, collection.content.src, **kwargs)
    return entry

  MoveResource = move_resource

  def delete_resource(self, entry, permanent=False, **kwargs):
    """Trashes or deletes the given entry.

    Args:
      entry: gdata.docs.data.Resource to trash or delete.
      permanent: True to skip the trash and delete the entry forever.
      kwargs: Other args to pass to gdata.client.GDClient.Delete()

    Returns:
      Result of delete request.
    """
    if permanent:
      kwargs['delete'] = 'true'
    return gdata.client.GDClient.delete(self, entry, **kwargs)

  DeleteResource = delete_resource

  def _check_entry_is_resource(self, entry):
    """Ensures given entry is a gdata.docs.data.Resource.

    Args:
      entry: Entry to test.
    Raises:
      ValueError: If given entry is not a resource.
    """
    if not isinstance(entry, gdata.docs.data.Resource):
      raise ValueError('%s is not a gdata.docs.data.Resource' % str(entry))

  def _check_entry_is_collection(self, entry):
    """Ensures given entry is a collection.

    Args:
      entry: Entry to test.
    Raises:
      ValueError: If given entry is a collection.
    """
    self._check_entry_is_resource(entry)
    if entry.get_resource_type() != gdata.docs.data.COLLECTION_LABEL:
      raise ValueError('%s is not a collection' % str(entry))

  def _check_entry_is_not_collection(self, entry):
    """Ensures given entry is not a collection.

    Args:
      entry: Entry to test.
    Raises:
      ValueError: If given entry is a collection.
    """
    try:
      self._check_entry_is_resource(entry)
    except ValueError:
      return
    if entry.get_resource_type() == gdata.docs.data.COLLECTION_LABEL:
      raise ValueError(
          '%s is a collection, which is not valid in this method' % str(entry))

  def _check_entry_has_content(self, entry):
    """Ensures given entry has a content element with src.

    Args:
      entry: Entry to test.
    Raises:
      ValueError: If given entry is not an entry or has no content src.
    """
    # Don't use _check_entry_is_resource since could be Revision etc
    if not (hasattr(entry, 'content') and entry.content and
            entry.content.src):
      raise ValueError('Entry %s has no downloadable content.' % entry)

  def get_acl(self, entry, **kwargs):
    """Retrieves an AclFeed for the given resource.

    Args:
      entry: gdata.docs.data.Resource to fetch AclFeed for.
      kwargs: Other args to pass to GetFeed().
    Returns:
      gdata.docs.data.AclFeed representing retrieved entries.
    """
    self._check_entry_is_resource(entry)
    return self.get_feed(
        entry.GetAclFeedLink().href,
        desired_class=gdata.docs.data.AclFeed, **kwargs)

  GetAcl = get_acl

  def get_acl_entry(self, entry, **kwargs):
    """Retrieves an AclEntry again.

    This is useful if you need to poll for an ACL changing.

    Args:
      entry: gdata.docs.data.AclEntry to fetch and return.
      kwargs: Other args to pass to GetAclEntryBySelfLink().
    Returns:
      gdata.docs.data.AclEntry representing retrieved entry.
    """

    return self.GetAclEntryBySelfLink(entry.GetSelfLink().href, **kwargs)

  GetAclEntry = get_acl_entry

  def get_acl_entry_by_self_link(self, self_link, **kwargs):
    """Retrieves a particular AclEntry by its self link.

    Args:
      self_link: URI at which to query for given ACL entry.  This can be found
          using entry.GetSelfLink().
      kwargs: Other parameters to pass to self.get_entry().

    Returns:
      gdata.docs.data.AclEntry representing the retrieved entry.
    """
    if isinstance(self_link, atom.data.Link):
      self_link = self_link.href

    return self.get_entry(self_link, desired_class=gdata.docs.data.AclEntry,
                          **kwargs)

  GetAclEntryBySelfLink = get_acl_entry_by_self_link

  def add_acl_entry(self, resource, acl_entry, send_notifications=None,
                    **kwargs):
    """Adds the given AclEntry to the given Resource.

    Args:
      resource: gdata.docs.data.Resource to which to add AclEntry.
      acl_entry: gdata.docs.data.AclEntry representing ACL entry to add.
      send_notifications: True if users should be notified by email when
          this AclEntry is added.
      kwargs: Other parameters to pass to self.post().

    Returns:
      gdata.docs.data.AclEntry containing information about new entry.
    Raises:
      ValueError: If given resource has no ACL link.
    """
    uri = resource.GetAclLink().href
    if uri is None:
      raise ValueError(('Given resource has no ACL link.  Did you fetch this'
                        'resource from the API?'))
    if send_notifications is not None:
      if not send_notifications:
        uri += '?send-notification-emails=false'

    return self.post(acl_entry, uri, desired_class=gdata.docs.data.AclEntry,
                     **kwargs)

  AddAclEntry = add_acl_entry

  def update_acl_entry(self, entry, send_notifications=None, **kwargs):
    """Updates the given AclEntry with new metadata.

    Args:
      entry: AclEntry to update. Make any metadata changes to this entry.
      send_notifications: True if users should be notified by email when
          this AclEntry is updated.
      kwargs: Other parameters to pass to super(DocsClient, self).update().

    Returns:
      gdata.docs.data.AclEntry representing the updated ACL entry.
    """
    uri = entry.GetEditLink().href
    if not send_notifications:
      uri += '?send-notification-emails=false'
    return super(DocsClient, self).update(entry, uri=uri, **kwargs)

  UpdateAclEntry = update_acl_entry

  def delete_acl_entry(self, entry, **kwargs):
    """Deletes the given AclEntry.

    Args:
      entry: gdata.docs.data.AclEntry to delete.
      kwargs: Other args to pass to gdata.client.GDClient.Delete()

    Returns:
      Result of delete request.
    """
    return super(DocsClient, self).delete(entry.GetEditLink().href,
                                          **kwargs)

  DeleteAclEntry = delete_acl_entry

  def batch_process_acl_entries(self, resource, entries, **kwargs):
    """Applies the specified operation of each entry in a single request.

    To use this, simply set acl_entry.batch_operation to one of
    ['query', 'insert', 'update', 'delete'], and optionally set
    acl_entry.batch_id to a string of your choice.

    Then, put all of your modified AclEntry objects into a list and pass
    that list as the entries parameter.

    Args:
      resource: gdata.docs.data.Resource to which the given entries belong.
      entries: [gdata.docs.data.AclEntry] to modify in some way.
      kwargs: Other args to pass to gdata.client.GDClient.post()

    Returns:
      Resulting gdata.docs.data.AclFeed of changes.
    """
    feed = gdata.docs.data.AclFeed()
    feed.entry = entries
    return super(DocsClient, self).post(
        feed, uri=resource.GetAclLink().href + '/batch', **kwargs)

  BatchProcessAclEntries = batch_process_acl_entries

  def get_revisions(self, entry, **kwargs):
    """Retrieves the revision history for a resource.

    Args:
      entry: gdata.docs.data.Resource for which to get revisions.
      kwargs: Other parameters to pass to self.get_feed().

    Returns:
      gdata.docs.data.RevisionFeed representing the resource's revisions.
    """
    self._check_entry_is_resource(entry)
    return self.get_feed(
        entry.GetRevisionsFeedLink().href,
        desired_class=gdata.docs.data.RevisionFeed, **kwargs)

  GetRevisions = get_revisions

  def get_revision(self, entry, **kwargs):
    """Retrieves a revision again given its entry.

    Args:
      entry: gdata.docs.data.Revision to fetch and return.
      kwargs: Other args to pass to GetRevisionBySelfLink().
    Returns:
      gdata.docs.data.Revision representing retrieved revision.
    """
    return self.GetRevisionBySelfLink(entry.GetSelfLink().href, **kwargs)

  GetRevision = get_revision

  def get_revision_by_self_link(self, self_link, **kwargs):
    """Retrieves a particular reivision by its self link.

    Args:
      self_link: URI at which to query for given revision.  This can be found
          using entry.GetSelfLink().
      kwargs: Other parameters to pass to self.get_entry().

    Returns:
      gdata.docs.data.Revision representing the retrieved revision.
    """
    if isinstance(self_link, atom.data.Link):
      self_link = self_link.href

    return self.get_entry(self_link, desired_class=gdata.docs.data.Revision,
                          **kwargs)

  GetRevisionBySelfLink = get_revision_by_self_link

  def download_revision(self, entry, file_path, extra_params=None, **kwargs):
    """Downloads the contents of the given revision to disk.

    Note: to download a revision in memory, use the DownloadRevisionToMemory()
    method.

    Args:
      entry: gdata.docs.data.Revision whose contents to fetch.
      file_path: str Full path to which to save file.
      extra_params: dict (optional) A map of any further parameters to control
          how the document is downloaded.
      kwargs: Other parameters to pass to self._download_file().

    Raises:
      gdata.client.RequestError if the download URL is malformed or the server's
      response was not successful.
    """
    self._check_entry_is_not_collection(entry)
    self._check_entry_has_content(entry)
    uri = self._get_download_uri(entry.content.src, extra_params)
    self._download_file(uri, file_path, **kwargs)

  DownloadRevision = download_revision

  def download_revision_to_memory(self, entry, extra_params=None, **kwargs):
    """Returns the contents of the given revision.

    Args:
      entry: gdata.docs.data.Revision whose contents to fetch.
      extra_params: dict (optional) A map of any further parameters to control
          how the document is downloaded/exported.
      kwargs: Other parameters to pass to self._get_content().

    Returns:
      Content of given revision after being downloaded.

    Raises:
      gdata.client.RequestError if the download URL is malformed or the server's
      response was not successful.
    """
    self._check_entry_is_not_collection(entry)
    self._check_entry_has_content(entry)
    uri = self._get_download_uri(entry.content.src, extra_params)
    return self._get_content(uri, **kwargs)

  DownloadRevisionToMemory = download_revision_to_memory

  def publish_revision(self, entry, publish_auto=None,
                       publish_outside_domain=False, **kwargs):
    """Publishes the given revision.

    This method can only be used for document revisions.

    Args:
      entry: Revision to update.
      publish_auto: True to automatically publish future revisions of the
          document.  False to not automatically publish future revisions.
          None to take no action and use the default value.
      publish_outside_domain: True to make the published revision available
          outside of a Google Apps domain.  False to not publish outside
          the domain.  None to use the default value.
      kwargs: Other parameters to pass to super(DocsClient, self).update().

    Returns:
      gdata.docs.data.Revision representing the updated revision.
    """
    entry.publish = gdata.docs.data.Publish(value='true')
    if publish_auto == True:
      entry.publish_auto = gdata.docs.data.PublishAuto(value='true')
    elif publish_auto == False:
      entry.publish_auto = gdata.docs.data.PublishAuto(value='false')
    if publish_outside_domain == True:
      entry.publish_outside_domain = \
          gdata.docs.data.PublishOutsideDomain(value='true')
    elif publish_outside_domain == False:
      entry.publish_outside_domain = \
          gdata.docs.data.PublishOutsideDomain(value='false')
    return super(DocsClient, self).update(entry, force=True, **kwargs)

  PublishRevision = publish_revision

  def unpublish_revision(self, entry, **kwargs):
    """Unpublishes the given revision.

    This method can only be used for document revisions.

    Args:
      entry: Revision to update.
      kwargs: Other parameters to pass to super(DocsClient, self).update().

    Returns:
      gdata.docs.data.Revision representing the updated revision.
    """
    entry.publish = gdata.docs.data.Publish(value='false')
    return super(DocsClient, self).update(entry, force=True, **kwargs)

  UnpublishRevision = unpublish_revision

  def delete_revision(self, entry, **kwargs):
    """Deletes the given Revision.

    Args:
      entry: gdata.docs.data.Revision to delete.
      kwargs: Other args to pass to gdata.client.GDClient.Delete()

    Returns:
      Result of delete request.
    """
    return super(DocsClient, self).delete(entry, force=True, **kwargs)

  DeleteRevision = delete_revision

  def get_archive(self, entry, **kwargs):
    """Retrieves an archive again given its entry.

    This is useful if you need to poll for an archive completing.

    Args:
      entry: gdata.docs.data.Archive to fetch and return.
      kwargs: Other args to pass to GetArchiveBySelfLink().
    Returns:
      gdata.docs.data.Archive representing retrieved archive.
    """

    return self.GetArchiveBySelfLink(entry.GetSelfLink().href, **kwargs)

  GetArchive = get_archive

  def get_archive_by_self_link(self, self_link, **kwargs):
    """Retrieves a particular archive by its self link.

    Args:
      self_link: URI at which to query for given archive.  This can be found
          using entry.GetSelfLink().
      kwargs: Other parameters to pass to self.get_entry().

    Returns:
      gdata.docs.data.Archive representing the retrieved archive.
    """
    if isinstance(self_link, atom.data.Link):
      self_link = self_link.href

    return self.get_entry(self_link, desired_class=gdata.docs.data.Archive,
                          **kwargs)

  GetArchiveBySelfLink = get_archive_by_self_link

  def create_archive(self, entry, **kwargs):
    """Creates a new archive of resources.

    Args:
      entry: gdata.docs.data.Archive representing metadata of archive to
          create.
      kwargs: Other parameters to pass to self.post().

    Returns:
      gdata.docs.data.Archive containing information about new archive.
    """
    return self.post(entry, ARCHIVE_FEED_URI,
                     desired_class=gdata.docs.data.Archive, **kwargs)

  CreateArchive = create_archive

  def update_archive(self, entry, **kwargs):
    """Updates the given Archive with new metadata.

    This method is really only useful for updating the notification email
    address of an archive that is being processed.

    Args:
      entry: Archive to update. Make any metadata changes to this entry.
      kwargs: Other parameters to pass to super(DocsClient, self).update().

    Returns:
      gdata.docs.data.Archive representing the updated archive.
    """
    return super(DocsClient, self).update(entry, **kwargs)

  UpdateArchive = update_archive

  download_archive = DownloadResource
  DownloadArchive = download_archive
  download_archive_to_memory = DownloadResourceToMemory
  DownloadArchiveToMemory = download_archive_to_memory

  def delete_archive(self, entry, **kwargs):
    """Aborts the given Archive operation, or deletes the Archive.

    Args:
      entry: gdata.docs.data.Archive to delete.
      kwargs: Other args to pass to gdata.client.GDClient.Delete()

    Returns:
      Result of delete request.
    """
    return super(DocsClient, self).delete(entry, force=True, **kwargs)

  DeleteArchive = delete_archive


class DocsQuery(gdata.client.Query):

  def __init__(self, title=None, title_exact=None, opened_min=None,
               opened_max=None, edited_min=None, edited_max=None, owner=None,
               writer=None, reader=None, show_collections=None, show_root=None,
               show_deleted=None, ocr=None, target_language=None,
               source_language=None, convert=None, query=None, **kwargs):
    """Constructs a query URL for the Google Documents List API.

    Args:
      title: str (optional) Specifies the search terms for the title of a
          document. This parameter used without title_exact will only
          submit partial queries, not exact queries.
      title_exact: str (optional) Meaningless without title. Possible values
          are 'true' and 'false'. Note: Matches are case-insensitive.
      opened_min: str (optional) Lower bound on the last time a document was
          opened by the current user. Use the RFC 3339 timestamp
          format. For example: opened_min='2005-08-09T09:57:00-08:00'.
      opened_max: str (optional) Upper bound on the last time a document was
          opened by the current user. (See also opened_min.)
      edited_min: str (optional) Lower bound on the last time a document was
          edited by the current user. This value corresponds to the edited.text
          value in the doc's entry object, which represents changes to the
          document's content or metadata.  Use the RFC 3339 timestamp format.
          For example: edited_min='2005-08-09T09:57:00-08:00'
      edited_max: str (optional) Upper bound on the last time a document was
          edited by the user. (See also edited_min.)
      owner: str (optional) Searches for documents with a specific owner. Use
          the email address of the owner. For example: owner='user@gmail.com'
      writer: str (optional) Searches for documents which can be written to
          by specific users. Use a single email address or a comma separated list
          of email addresses. For example: writer='user1@gmail.com,user@example.com'
      reader: str (optional) Searches for documents which can be read by
          specific users. (See also writer.)
      show_collections: str (optional) Specifies whether the query should return
          collections as well as documents and files. Possible values are 'true'
          and 'false'. Default is 'false'.
      show_root: (optional) 'true' to specify when an item is in the root
          collection. Default is 'false'
      show_deleted: str (optional) Specifies whether the query should return
          documents which are in the trash as well as other documents.
          Possible values are 'true' and 'false'. Default is false.
      ocr: str (optional) Specifies whether to attempt OCR on a .jpg, .png, or
          .gif upload. Possible values are 'true' and 'false'. Default is
          false. See OCR in the Protocol Guide:
          http://code.google.com/apis/documents/docs/3.0/developers_guide_protocol.html#OCR
      target_language: str (optional) Specifies the language to translate a
          document into. See Document Translation in the Protocol Guide for a
          table of possible values:
            http://code.google.com/apis/documents/docs/3.0/developers_guide_protocol.html#DocumentTranslation
      source_language: str (optional) Specifies the source language of the
          original document. Optional when using the translation service.
          If not provided, Google will attempt to auto-detect the source
          language. See Document Translation in the Protocol Guide for a table of
          possible values (link in target_language).
      convert: str (optional) Used when uploading files specify if document uploads
          should convert to a native Google Docs format.
          Possible values are 'true' and 'false'. The default is 'true'.
      query: str (optional) Full-text query to use.  See the 'q' parameter in
          the documentation.
    """
    gdata.client.Query.__init__(self, **kwargs)
    self.convert = convert
    self.title = title
    self.title_exact = title_exact
    self.opened_min = opened_min
    self.opened_max = opened_max
    self.edited_min = edited_min
    self.edited_max = edited_max
    self.owner = owner
    self.writer = writer
    self.reader = reader
    self.show_collections = show_collections
    self.show_root = show_root
    self.show_deleted = show_deleted
    self.ocr = ocr
    self.target_language = target_language
    self.source_language = source_language
    self.query = query

  def modify_request(self, http_request):
    gdata.client._add_query_param('convert', self.convert, http_request)
    gdata.client._add_query_param('title', self.title, http_request)
    gdata.client._add_query_param('title-exact', self.title_exact,
                                  http_request)
    gdata.client._add_query_param('opened-min', self.opened_min, http_request)
    gdata.client._add_query_param('opened-max', self.opened_max, http_request)
    gdata.client._add_query_param('edited-min', self.edited_min, http_request)
    gdata.client._add_query_param('edited-max', self.edited_max, http_request)
    gdata.client._add_query_param('owner', self.owner, http_request)
    gdata.client._add_query_param('writer', self.writer, http_request)
    gdata.client._add_query_param('reader', self.reader, http_request)
    gdata.client._add_query_param('query', self.query, http_request)
    gdata.client._add_query_param('showfolders', self.show_collections,
                                  http_request)
    gdata.client._add_query_param('showroot', self.show_root, http_request)
    gdata.client._add_query_param('showdeleted', self.show_deleted,
                                  http_request)
    gdata.client._add_query_param('ocr', self.ocr, http_request)
    gdata.client._add_query_param('targetLanguage', self.target_language,
                                  http_request)
    gdata.client._add_query_param('sourceLanguage', self.source_language,
                                  http_request)
    gdata.client.Query.modify_request(self, http_request)

  ModifyRequest = modify_request



from collections import Iterable
from functools import partial, wraps

from debug import get_name, set_name, repr_args
from compat import text_type, string_types, dict_items


class Pipe(object):
    """
    Pipe-style combinator.

    Example::

        p = pipe | F | G | H

        p(x) == H(G(F(x)))

    """
    def __init__(self, func=None):
        self.func = func
        self.__name__ = 'Pipe'

    def __str__(self):
        return get_name(self.func)

    __repr__ = __str__

    @staticmethod
    def compose(first, second):
        name = lambda: '{0} | {1}'.format(get_name(first), get_name(second))

        def composite(*args, **kwargs):
            return second(first(*args, **kwargs))
        return set_name(name, composite)

    @classmethod
    def bind(cls, first, second):
        return cls(
            first if second is None else
            second if first is None else
            cls.compose(first, second))

    def __or__(self, next_func):
        return self.bind(self.func, prepare_function_for_pipe(next_func))

    def __ror__(self, prev_func):
        return self.bind(prepare_function_for_pipe(prev_func), self.func)

    def __lt__(self, thing):
        return self.func(thing) if self.func else thing

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __get__(self, instance, owner):
        return partial(self, instance) if instance else self


pipe = Pipe()


class Maybe(Pipe):

    @staticmethod
    def compose(first, second):
        name = lambda: '{0} ?| {1}'.format(get_name(first), get_name(second))

        def composite(*args, **kwargs):
            result = first(*args, **kwargs)
            return None if result is None else second(result)
        return set_name(name, composite)

    def __call__(self, *args, **kwargs):
        if len(args) == 1 and args[0] is None and not kwargs:
            return None
        return self.func(*args, **kwargs)

    def __lt__(self, thing):
        return (
            None if thing is None else
            self.func(thing) if self.func else
            thing)

maybe = Maybe()


def prepare_function_for_pipe(thing):
    if isinstance(thing, XObject):
        return ~thing
    if isinstance(thing, tuple):
        return xpartial(*thing)
    if isinstance(thing, string_types):
        return StringFormatter(thing)
    if callable(thing):
        return thing
    raise ValueError('Cannot pipe %s' % thing)


def StringFormatter(template):

    f = text_type(template).format

    def format(content):
        if isinstance(content, dict):
            return f(**content)
        if _iterable(content):
            return f(*content)
        return f(content)

    return set_name(lambda: "format('%s')" % template[:20], format)


def _iterable(obj):
    "Iterable but not a string"
    return isinstance(obj, Iterable) and not isinstance(obj, string_types)


class XObject(object):

    def __init__(self, func=None):
        self._func = func
        set_name(lambda: get_name(func) if func else 'X', self)

    def __repr__(self):
        return get_name(self)

    def __invert__(self):
        return self._func or set_name('X', lambda x: x)

    def bind(self, name, func):
        set_name(name, func)
        return XObject((self._func | func) if self._func else (pipe | func))

    def __call__(self, *args, **kwargs):
        name = lambda: 'X(%s)' % repr_args(*args, **kwargs)
        return self.bind(name, lambda x: x(*args, **kwargs))

    def __hash__(self):
        return super(XObject, self).__hash__()

    def __eq__(self, other):
        return self.bind(lambda: 'X == {0!r}'.format(other), lambda x: x == other)

    def __getattr__(self, name):
        return self.bind(lambda: 'X.{0}'.format(name), lambda x: getattr(x, name))

    def __getitem__(self, item):
        return self.bind(lambda: 'X[{0!r}]'.format(item), lambda x: x[item])

    def __gt__(self, other):
        return self.bind(lambda: 'X > {0!r}'.format(other), lambda x: x > other)

    def __ge__(self, other):
        return self.bind(lambda: 'X >= {0!r}'.format(other), lambda x: x >= other)

    def __lt__(self, other):
        return self.bind(lambda: 'X < {0!r}'.format(other), lambda x: x < other)

    def __le__(self, other):
        return self.bind(lambda: 'X <= {0!r}'.format(other), lambda x: x <= other)

    def __mod__(self, y):
        return self.bind(lambda: 'X % {0!r}'.format(y), lambda x: x % y)

    def __ne__(self, other):
        return self.bind(lambda: 'X != {0!r}'.format(other), lambda x: x != other)

    def __neg__(self):
        return self.bind(lambda: '-X', lambda x: -x)

    def __mul__(self, other):
        return self.bind(lambda: 'X * {0!r}'.format(other), lambda x: x * other)

    def __div__(self, other):
        return self.bind(lambda: 'X / {0!r}'.format(other), lambda x: x / other)

    def __add__(self, other):
        return self.bind(lambda: 'X + {0!r}'.format(other), lambda x: x + other)

    def __sub__(self, other):
        return self.bind(lambda: 'X - {0!r}'.format(other), lambda x: x - other)

    def __pow__(self, other):
        return self.bind(lambda: 'X ** {0!r}'.format(other), lambda x: x ** other)

    def __ror__(self, func):
        return pipe | func | self

    def __or__(self, func):
        if isinstance(func, Pipe):
            return func.__ror__(self)
        return pipe | self | func

    def _in_(self, y):
        return self.bind(lambda: 'X._in_({0!r})'.format(y), lambda x: x in y)


X = XObject()


def xpartial(func, *xargs, **xkwargs):
    """
    Like :func:`functools.partial`, but can take an :class:`XObject`
    placeholder that will be replaced with the first positional argument
    when the partially applied function is called.

    Useful when the function's positional arguments' order doesn't fit your
    situation, e.g.:

    >>> reverse_range = xpartial(range, X, 0, -1)
    >>> reverse_range(5)
    [5, 4, 3, 2, 1]

    It can also be used to transform the positional argument to a keyword
    argument, which can come in handy inside a *pipe*::

        xpartial(objects.get, id=X)

    Also the XObjects are evaluated, which can be used for some sort of
    destructuring of the argument::

        xpartial(somefunc, name=X.name, number=X.contacts['number'])

    Lastly, unlike :func:`functools.partial`, this creates a regular function
    which will bind to classes (like the ``curry`` function from
    ``django.utils.functional``).
    """
    any_x = any(isinstance(a, XObject) for a in xargs + tuple(xkwargs.values()))
    use = lambda x, value: (~x)(value) if isinstance(x, XObject) else x

    @wraps(func)
    def xpartially_applied(*func_args, **func_kwargs):
        if any_x:
            if not func_args:
                raise ValueError('Function "%s" partially applied with an '
                    'X placeholder but called with no positional arguments.'
                    % get_name(func))
            first = func_args[0]
            rest = func_args[1:]
            args = tuple(use(x, first) for x in xargs) + rest
            kwargs = dict((k, use(x, first)) for k, x in dict_items(xkwargs))
            kwargs.update(func_kwargs)
        else:
            args = xargs + func_args
            kwargs = dict(xkwargs, **func_kwargs)
        return func(*args, **kwargs)

    name = lambda: '%s(%s)' % (get_name(func), repr_args(*xargs, **xkwargs))
    return set_name(name, xpartially_applied)



#!/usr/bin/python
#
# Original Copyright (C) 2006 Google Inc.
# Refactored in 2009 to work for Google Analytics by Sal Uryasev at Juice Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Note that this module will not function without specifically adding
# 'analytics': [  #Google Analytics
#  'https://www.google.com/analytics/feeds/'],
# to CLIENT_LOGIN_SCOPES in the gdata/service.py file

"""Contains extensions to Atom objects used with Google Analytics."""

__author__ = 'api.suryasev (Sal Uryasev)'

import atom
import gdata

GAN_NAMESPACE = 'http://schemas.google.com/analytics/2009'

class TableId(gdata.GDataEntry):
  """tableId element."""
  _tag = 'tableId'
  _namespace = GAN_NAMESPACE  

class Property(gdata.GDataEntry):
  _tag = 'property'
  _namespace = GAN_NAMESPACE
  _children = gdata.GDataEntry._children.copy()
  _attributes = gdata.GDataEntry._attributes.copy()
  
  _attributes['name'] = 'name'
  _attributes['value'] = 'value'
  
  def __init__(self, name=None, value=None, *args, **kwargs):
    self.name = name
    self.value = value
    super(Property, self).__init__(*args, **kwargs)
    
  def __str__(self):
    return self.value
      
  def __repr__(self):
    return self.value

class AccountListEntry(gdata.GDataEntry):
  """The Google Documents version of an Atom Entry"""
  
  _tag = 'entry'
  _namespace = atom.ATOM_NAMESPACE
  _children = gdata.GDataEntry._children.copy()
  _attributes = gdata.GDataEntry._attributes.copy()
  _children['{%s}tableId' % GAN_NAMESPACE] = ('tableId', 
                          [TableId])
  _children['{%s}property' % GAN_NAMESPACE] = ('property', 
                          [Property])

  def __init__(self, tableId=None, property=None,
         *args, **kwargs):
    self.tableId = tableId
    self.property = property
    super(AccountListEntry, self).__init__(*args, **kwargs)


def AccountListEntryFromString(xml_string):
  """Converts an XML string into an AccountListEntry object.

  Args:
  xml_string: string The XML describing a Document List feed entry.

  Returns:
  A AccountListEntry object corresponding to the given XML.
  """
  return atom.CreateClassFromXMLString(AccountListEntry, xml_string)


class AccountListFeed(gdata.GDataFeed):
  """A feed containing a list of Google Documents Items"""
  
  _tag = 'feed'
  _namespace = atom.ATOM_NAMESPACE
  _children = gdata.GDataFeed._children.copy()
  _attributes = gdata.GDataFeed._attributes.copy()
  _children['{%s}entry' % atom.ATOM_NAMESPACE] = ('entry', 
                          [AccountListEntry])


def AccountListFeedFromString(xml_string):
  """Converts an XML string into an AccountListFeed object.

  Args:
  xml_string: string The XML describing an AccountList feed.

  Returns:
  An AccountListFeed object corresponding to the given XML.
  All properties are also linked to with a direct reference
  from each entry object for convenience. (e.g. entry.AccountName)
  """
  feed = atom.CreateClassFromXMLString(AccountListFeed, xml_string)
  for entry in feed.entry:
    for pro in entry.property:
      entry.__dict__[pro.name.replace('ga:','')] = pro
    for td in entry.tableId:
      td.__dict__['value'] = td.text
  return feed

class Dimension(gdata.GDataEntry):
  _tag = 'dimension'
  _namespace = GAN_NAMESPACE
  _children = gdata.GDataEntry._children.copy()
  _attributes = gdata.GDataEntry._attributes.copy()
  
  _attributes['name'] = 'name'
  _attributes['value'] = 'value'
  _attributes['type'] = 'type'
  _attributes['confidenceInterval'] = 'confidence_interval'
  
  def __init__(self, name=None, value=None, type=None, 
               confidence_interval = None, *args, **kwargs):
    self.name = name
    self.value = value
    self.type = type
    self.confidence_interval = confidence_interval
    super(Dimension, self).__init__(*args, **kwargs)   
    
  def __str__(self):
    return self.value

  def __repr__(self):
    return self.value 
    
class Metric(gdata.GDataEntry):
  _tag = 'metric'
  _namespace = GAN_NAMESPACE
  _children = gdata.GDataEntry._children.copy()
  _attributes = gdata.GDataEntry._attributes.copy()
  
  _attributes['name'] = 'name'
  _attributes['value'] = 'value'
  _attributes['type'] = 'type'
  _attributes['confidenceInterval'] = 'confidence_interval'
  
  def __init__(self, name=None, value=None, type=None, 
               confidence_interval = None, *args, **kwargs):
    self.name = name
    self.value = value
    self.type = type
    self.confidence_interval = confidence_interval
    super(Metric, self).__init__(*args, **kwargs)
    
  def __str__(self):
    return self.value

  def __repr__(self):
    return self.value
  
class AnalyticsDataEntry(gdata.GDataEntry):
  """The Google Analytics version of an Atom Entry"""
  
  _tag = 'entry'
  _namespace = atom.ATOM_NAMESPACE
  _children = gdata.GDataEntry._children.copy()
  _attributes = gdata.GDataEntry._attributes.copy()
  
  _children['{%s}dimension' % GAN_NAMESPACE] = ('dimension', 
                          [Dimension])
                          
  _children['{%s}metric' % GAN_NAMESPACE] = ('metric', 
                         [Metric])
                         
  def __init__(self, dimension=None, metric=None, *args, **kwargs):
    self.dimension = dimension
    self.metric = metric
    
    super(AnalyticsDataEntry, self).__init__(*args, **kwargs)

class AnalyticsDataFeed(gdata.GDataFeed):
  """A feed containing a list of Google Analytics Data Feed"""
  
  _tag = 'feed'
  _namespace = atom.ATOM_NAMESPACE
  _children = gdata.GDataFeed._children.copy()
  _attributes = gdata.GDataFeed._attributes.copy()
  _children['{%s}entry' % atom.ATOM_NAMESPACE] = ('entry', 
                          [AnalyticsDataEntry])
                    
                         
"""
Data Feed
"""

def AnalyticsDataFeedFromString(xml_string):
  """Converts an XML string into an AccountListFeed object.

  Args:
  xml_string: string The XML describing an AccountList feed.

  Returns:
  An AccountListFeed object corresponding to the given XML.
  Each metric and dimension is also referenced directly from
  the entry for easier access. (e.g. entry.keyword.value)
  """
  feed = atom.CreateClassFromXMLString(AnalyticsDataFeed, xml_string)
  if feed.entry:
    for entry in feed.entry:
      for met in entry.metric:
        entry.__dict__[met.name.replace('ga:','')] = met
      if entry.dimension is not None:
        for dim in entry.dimension:
          entry.__dict__[dim.name.replace('ga:','')] = dim
        
  return feed



#!/usr/bin/python
#
# Copyright (C) 2008 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides a TokenStore class which is designed to manage
auth tokens required for different services.

Each token is valid for a set of scopes which is the start of a URL. An HTTP
client will use a token store to find a valid Authorization header to send
in requests to the specified URL. If the HTTP client determines that a token
has expired or been revoked, it can remove the token from the store so that
it will not be used in future requests.
"""


__author__ = 'api.jscudder (Jeff Scudder)'


import atom.http_interface
import atom.url


SCOPE_ALL = 'http'


class TokenStore(object):
  """Manages Authorization tokens which will be sent in HTTP headers."""
  def __init__(self, scoped_tokens=None):
    self._tokens = scoped_tokens or {}

  def add_token(self, token):
    """Adds a new token to the store (replaces tokens with the same scope).

    Args:
      token: A subclass of http_interface.GenericToken. The token object is 
          responsible for adding the Authorization header to the HTTP request.
          The scopes defined in the token are used to determine if the token
          is valid for a requested scope when find_token is called.

    Returns:
      True if the token was added, False if the token was not added becase
      no scopes were provided.
    """
    if not hasattr(token, 'scopes') or not token.scopes:
      return False

    for scope in token.scopes:
      self._tokens[str(scope)] = token
    return True  

  def find_token(self, url):
    """Selects an Authorization header token which can be used for the URL.

    Args:
      url: str or atom.url.Url or a list containing the same.
          The URL which is going to be requested. All
          tokens are examined to see if any scopes begin match the beginning
          of the URL. The first match found is returned.

    Returns:
      The token object which should execute the HTTP request. If there was
      no token for the url (the url did not begin with any of the token
      scopes available), then the atom.http_interface.GenericToken will be 
      returned because the GenericToken calls through to the http client
      without adding an Authorization header.
    """
    if url is None:
      return None
    if isinstance(url, (str, unicode)):
      url = atom.url.parse_url(url)
    if url in self._tokens:
      token = self._tokens[url]
      if token.valid_for_scope(url):
        return token
      else:
        del self._tokens[url]
    for scope, token in self._tokens.iteritems():
      if token.valid_for_scope(url):
        return token
    return atom.http_interface.GenericToken()

  def remove_token(self, token):
    """Removes the token from the token_store.

    This method is used when a token is determined to be invalid. If the
    token was found by find_token, but resulted in a 401 or 403 error stating
    that the token was invlid, then the token should be removed to prevent
    future use.

    Returns:
      True if a token was found and then removed from the token
      store. False if the token was not in the TokenStore.
    """
    token_found = False
    scopes_to_delete = []
    for scope, stored_token in self._tokens.iteritems():
      if stored_token == token:
        scopes_to_delete.append(scope)
        token_found = True
    for scope in scopes_to_delete:
      del self._tokens[scope]
    return token_found

  def remove_all_tokens(self):
    self._tokens = {} 



"""This file implements all-or-nothing package transformations.

An all-or-nothing package transformation is one in which some text is
transformed into message blocks, such that all blocks must be obtained before
the reverse transformation can be applied.  Thus, if any blocks are corrupted
or lost, the original message cannot be reproduced.

An all-or-nothing package transformation is not encryption, although a block
cipher algorithm is used.  The encryption key is randomly generated and is
extractable from the message blocks.

This class implements the All-Or-Nothing package transformation algorithm
described in:

Ronald L. Rivest.  "All-Or-Nothing Encryption and The Package Transform"
http://theory.lcs.mit.edu/~rivest/fusion.pdf

"""

__revision__ = "$Id: AllOrNothing.py,v 1.8 2003/02/28 15:23:20 akuchling Exp $"

import operator
import string
from Crypto.Util.number import bytes_to_long, long_to_bytes



class AllOrNothing:
    """Class implementing the All-or-Nothing package transform.

    Methods for subclassing:

        _inventkey(key_size):
            Returns a randomly generated key.  Subclasses can use this to
            implement better random key generating algorithms.  The default
            algorithm is probably not very cryptographically secure.

    """

    def __init__(self, ciphermodule, mode=None, IV=None):
        """AllOrNothing(ciphermodule, mode=None, IV=None)

        ciphermodule is a module implementing the cipher algorithm to
        use.  It must provide the PEP272 interface.

        Note that the encryption key is randomly generated
        automatically when needed.  Optional arguments mode and IV are
        passed directly through to the ciphermodule.new() method; they
        are the feedback mode and initialization vector to use.  All
        three arguments must be the same for the object used to create
        the digest, and to undigest'ify the message blocks.
        """

        self.__ciphermodule = ciphermodule
        self.__mode = mode
        self.__IV = IV
        self.__key_size = ciphermodule.key_size
        if self.__key_size == 0:
            self.__key_size = 16

    __K0digit = chr(0x69)

    def digest(self, text):
        """digest(text:string) : [string]

        Perform the All-or-Nothing package transform on the given
        string.  Output is a list of message blocks describing the
        transformed text, where each block is a string of bit length equal
        to the ciphermodule's block_size.
        """

        # generate a random session key and K0, the key used to encrypt the
        # hash blocks.  Rivest calls this a fixed, publically-known encryption
        # key, but says nothing about the security implications of this key or
        # how to choose it.
        key = self._inventkey(self.__key_size)
        K0 = self.__K0digit * self.__key_size

        # we need two cipher objects here, one that is used to encrypt the
        # message blocks and one that is used to encrypt the hashes.  The
        # former uses the randomly generated key, while the latter uses the
        # well-known key.
        mcipher = self.__newcipher(key)
        hcipher = self.__newcipher(K0)

        # Pad the text so that its length is a multiple of the cipher's
        # block_size.  Pad with trailing spaces, which will be eliminated in
        # the undigest() step.
        block_size = self.__ciphermodule.block_size
        padbytes = block_size - (len(text) % block_size)
        text = text + ' ' * padbytes

        # Run through the algorithm:
        # s: number of message blocks (size of text / block_size)
        # input sequence: m1, m2, ... ms
        # random key K' (`key' in the code)
        # Compute output sequence: m'1, m'2, ... m's' for s' = s + 1
        # Let m'i = mi ^ E(K', i) for i = 1, 2, 3, ..., s
        # Let m's' = K' ^ h1 ^ h2 ^ ... hs
        # where hi = E(K0, m'i ^ i) for i = 1, 2, ... s
        #
        # The one complication I add is that the last message block is hard
        # coded to the number of padbytes added, so that these can be stripped
        # during the undigest() step
        s = len(text) / block_size
        blocks = []
        hashes = []
        for i in range(1, s+1):
            start = (i-1) * block_size
            end = start + block_size
            mi = text[start:end]
            assert len(mi) == block_size
            cipherblock = mcipher.encrypt(long_to_bytes(i, block_size))
            mticki = bytes_to_long(mi) ^ bytes_to_long(cipherblock)
            blocks.append(mticki)
            # calculate the hash block for this block
            hi = hcipher.encrypt(long_to_bytes(mticki ^ i, block_size))
            hashes.append(bytes_to_long(hi))

        # Add the padbytes length as a message block
        i = i + 1
        cipherblock = mcipher.encrypt(long_to_bytes(i, block_size))
        mticki = padbytes ^ bytes_to_long(cipherblock)
        blocks.append(mticki)

        # calculate this block's hash
        hi = hcipher.encrypt(long_to_bytes(mticki ^ i, block_size))
        hashes.append(bytes_to_long(hi))

        # Now calculate the last message block of the sequence 1..s'.  This
        # will contain the random session key XOR'd with all the hash blocks,
        # so that for undigest(), once all the hash blocks are calculated, the
        # session key can be trivially extracted.  Calculating all the hash
        # blocks requires that all the message blocks be received, thus the
        # All-or-Nothing algorithm succeeds.
        mtick_stick = bytes_to_long(key) ^ reduce(operator.xor, hashes)
        blocks.append(mtick_stick)

        # we convert the blocks to strings since in Python, byte sequences are
        # always represented as strings.  This is more consistent with the
        # model that encryption and hash algorithms always operate on strings.
        return map(long_to_bytes, blocks)


    def undigest(self, blocks):
        """undigest(blocks : [string]) : string

        Perform the reverse package transformation on a list of message
        blocks.  Note that the ciphermodule used for both transformations
        must be the same.  blocks is a list of strings of bit length
        equal to the ciphermodule's block_size.
        """

        # better have at least 2 blocks, for the padbytes package and the hash
        # block accumulator
        if len(blocks) < 2:
            raise ValueError, "List must be at least length 2."

        # blocks is a list of strings.  We need to deal with them as long
        # integers
        blocks = map(bytes_to_long, blocks)

        # Calculate the well-known key, to which the hash blocks are
        # encrypted, and create the hash cipher.
        K0 = self.__K0digit * self.__key_size
        hcipher = self.__newcipher(K0)

        # Since we have all the blocks (or this method would have been called
        # prematurely), we can calcualte all the hash blocks.
        hashes = []
        for i in range(1, len(blocks)):
            mticki = blocks[i-1] ^ i
            hi = hcipher.encrypt(long_to_bytes(mticki))
            hashes.append(bytes_to_long(hi))

        # now we can calculate K' (key).  remember the last block contains
        # m's' which we don't include here
        key = blocks[-1] ^ reduce(operator.xor, hashes)

        # and now we can create the cipher object
        mcipher = self.__newcipher(long_to_bytes(key))
        block_size = self.__ciphermodule.block_size

        # And we can now decode the original message blocks
        parts = []
        for i in range(1, len(blocks)):
            cipherblock = mcipher.encrypt(long_to_bytes(i, block_size))
            mi = blocks[i-1] ^ bytes_to_long(cipherblock)
            parts.append(mi)

        # The last message block contains the number of pad bytes appended to
        # the original text string, such that its length was an even multiple
        # of the cipher's block_size.  This number should be small enough that
        # the conversion from long integer to integer should never overflow
        padbytes = int(parts[-1])
        text = string.join(map(long_to_bytes, parts[:-1]), '')
        return text[:-padbytes]

    def _inventkey(self, key_size):
        # TBD: Not a very secure algorithm.  Eventually, I'd like to use JHy's
        # kernelrand module
        import time
        from Crypto.Util import randpool
        # TBD: key_size * 2 to work around possible bug in RandomPool?
        pool = randpool.RandomPool(key_size * 2)
        while key_size > pool.entropy:
            pool.add_event()

        # we now have enough entropy in the pool to get a key_size'd key
        return pool.get_bytes(key_size)

    def __newcipher(self, key):
        if self.__mode is None and self.__IV is None:
            return self.__ciphermodule.new(key)
        elif self.__IV is None:
            return self.__ciphermodule.new(key, self.__mode)
        else:
            return self.__ciphermodule.new(key, self.__mode, self.__IV)



if __name__ == '__main__':
    import sys
    import getopt
    import base64

    usagemsg = '''\
Test module usage: %(program)s [-c cipher] [-l] [-h]

Where:
    --cipher module
    -c module
        Cipher module to use.  Default: %(ciphermodule)s

    --aslong
    -l
        Print the encoded message blocks as long integers instead of base64
        encoded strings

    --help
    -h
        Print this help message
'''

    ciphermodule = 'AES'
    aslong = 0

    def usage(code, msg=None):
        if msg:
            print msg
        print usagemsg % {'program': sys.argv[0],
                          'ciphermodule': ciphermodule}
        sys.exit(code)

    try:
        opts, args = getopt.getopt(sys.argv[1:],
                                   'c:l', ['cipher=', 'aslong'])
    except getopt.error, msg:
        usage(1, msg)

    if args:
        usage(1, 'Too many arguments')

    for opt, arg in opts:
        if opt in ('-h', '--help'):
            usage(0)
        elif opt in ('-c', '--cipher'):
            ciphermodule = arg
        elif opt in ('-l', '--aslong'):
            aslong = 1

    # ugly hack to force __import__ to give us the end-path module
    module = __import__('Crypto.Cipher.'+ciphermodule, None, None, ['new'])

    a = AllOrNothing(module)
    print 'Original text:\n=========='
    print __doc__
    print '=========='
    msgblocks = a.digest(__doc__)
    print 'message blocks:'
    for i, blk in map(None, range(len(msgblocks)), msgblocks):
        # base64 adds a trailing newline
        print '    %3d' % i,
        if aslong:
            print bytes_to_long(blk)
        else:
            print base64.encodestring(blk)[:-1]
    #
    # get a new undigest-only object so there's no leakage
    b = AllOrNothing(module)
    text = b.undigest(msgblocks)
    if text == __doc__:
        print 'They match!'
    else:
        print 'They differ!'
